@@ -11,7 +11,8 @@ test_that("no deadlock when no stdout + wait", {
 
 test_that("wait with timeout", {
 
-  p <- process$new(commandline = sleep(3))
+  px <- get_tool("px")
+  p <- process$new(px, c("sleep", "3"))
   expect_true(p$is_alive())
 
   t1 <- proc.time()
@@ -28,12 +29,12 @@ test_that("wait with timeout", {
 
 test_that("wait after process already exited", {
 
-  cmd <- if (os_type() == "windows") "dir /b /A" else "ls -A"
-  p <- process$new(commandline = cmd)
+  px <- get_tool("px")
+  p <- process$new(
+    px, c("outln", "foo", "outln", "bar", "outln", "foobar"))
 
-  ## Make sure it is done, wait a bit, so that exit status is collected
+  ## Make sure it is done
   p$wait()
-  Sys.sleep(1)
 
   ## Now wait() should return immediately, regardless of timeout
   expect_true(system.time(p$wait())[["elapsed"]] < 1)
